/*
* SACD Decoder plugin
* Copyright (c) 2011-2015 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "audio_config.h"
#include "asio_device.h"

static const LPSTR ASIO_PATH = "SOFTWARE\\ASIO";
static const LPSTR ASIO_CLSID = "CLSID";

void asio_device_t::enumerate(output_device_enum_callback& p_callback) {
	LONG rc;
	HKEY hkAsio;
	rc = RegOpenKeyExA(HKEY_LOCAL_MACHINE, ASIO_PATH, 0, KEY_READ, &hkAsio);
	if (rc == ERROR_SUCCESS) {
		DWORD nDrivers = 0;
		DWORD cMaxDriverNameLen;
		rc = RegQueryInfoKeyA(hkAsio, NULL, NULL, NULL, &nDrivers, &cMaxDriverNameLen, NULL, NULL, NULL, NULL, NULL, NULL);
		if (rc == ERROR_SUCCESS) {
			char* lpszDriverName = new char[cMaxDriverNameLen + 1];
			if (lpszDriverName) {
				for (DWORD dwIndex = 0; dwIndex < nDrivers; dwIndex++) {
					DWORD cDriverNameLen = cMaxDriverNameLen + 1;
					rc = RegEnumKeyExA(hkAsio, dwIndex, (LPSTR) lpszDriverName, &cDriverNameLen, NULL, NULL, NULL, NULL);
					if (rc == ERROR_SUCCESS) {
						HKEY hkAsioDriver;
						CLSID clsid;
						DWORD dwType;
						DWORD cbData;
						rc = RegOpenKeyExA(hkAsio, (LPSTR) lpszDriverName, 0, KEY_READ, &hkAsioDriver);
						if (rc == ERROR_SUCCESS) {
							rc = RegQueryValueExA(hkAsioDriver, ASIO_CLSID, NULL, &dwType, NULL, &cbData);
							if (rc == ERROR_SUCCESS) {
								char* lpszCLSID = new char[cbData];
								if (lpszCLSID) {
									rc = RegQueryValueExA(hkAsioDriver, ASIO_CLSID, NULL, &dwType, (LPBYTE) lpszCLSID, &cbData);
									if (rc == ERROR_SUCCESS) {
										clsid = pfc::GUID_from_text(lpszCLSID);
										pfc::string8 name;
										name << "ASIO : " << lpszDriverName;
										p_callback.on_device(clsid, name.c_str(), name.length());
									}
									delete [] lpszCLSID;
								}
							}
							RegCloseKey(hkAsioDriver);
						}
					}
					if (rc == ERROR_NO_MORE_ITEMS) {
						break;
					}
				}
				delete [] lpszDriverName;
			}
		}
		RegCloseKey(hkAsio);
	}
}

asio_device_t::asio_device_t() {
	InitializeCriticalSection(&m_cs);
	m_guid = GUID_NULL;

	m_pCoASIO = nullptr;
	m_state = ASIO_STATE_UNLOADED;
	m_channels = 0;
	m_samples = 0;
	m_can_DSD = m_run_DSD = false;
	m_name[0] = '\0';
	m_version = 0;
	m_numInputChannels = 0;
	m_numOutputChannels = 0;
	m_minSize = 0;
	m_maxSize = 0;
	m_preferredSize = 0;
	m_granularity = 0;
	m_inputLatency = 0;
	m_outputLatency = 0;
}

asio_device_t::~asio_device_t() {
	DeleteCriticalSection(&m_cs);
}

bool asio_device_t::load(const GUID& p_GUID) {
	HRESULT hr;
	bool ok = false;
	EnterCriticalSection(&m_cs);
	m_guid = p_GUID;
	m_trace = audio_config_t::g_get_trace();
	if (!m_pCoASIO) {
		hr = CoCreateInstance(m_guid, nullptr, CLSCTX_INPROC_SERVER, m_guid, (LPVOID*) &m_pCoASIO);
		if (hr == S_OK) {
			m_state = ASIO_STATE_LOADED;
			if (init(core_api::get_main_window())) {
				m_can_DSD = can_do_format(kASIODSDFormat);
				if (set_channels()) {
					if (set_channelInfos()) {
						ok = true;
					}
				}
			}
			if (m_trace) {
				console::printf(TRC_INF "Driver is loaded [name = \"%s\", state = %d]", m_name, m_state);
			}
		}
		else {
			console::printf(TRC_ERR "CoCreateInstance() => 0x%08x", hr);
		}
	}
	else {
		console::printf(TRC_ERR "Driver is already loaded");
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::unload() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	if (m_pCoASIO->Release() > 0) {
		m_state = ASIO_STATE_LOADED;
		if (m_trace) {
			console::printf(TRC_INF "Driver is released [name = \"%s\", state = %d]", m_name, m_state);
		}
	}
	else {
		m_pCoASIO = nullptr;
		m_state = ASIO_STATE_UNLOADED;
		if (m_trace) {
			console::printf(TRC_INF "Driver is unloaded [name = \"%s\", state = %d]", m_name, m_state);
		}
		ok = true;
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::configure() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->controlPanel();
	if (rc == ASE_OK) {
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::controlPanel() => %s", asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::can_time_info() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	if (future(kAsioCanTimeInfo, nullptr)) {
		ok = true;
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::can_report_overload() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	if (future(kAsioCanReportOverload, nullptr)) {
		ok = true;
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::can_do_format(ASIOIoFormatType p_format) {
	bool ok = false;
	ASIOIoFormat opt;
	opt.FormatType = p_format;
	EnterCriticalSection(&m_cs);
	if (future(kAsioCanDoIoFormat, &opt)) {
		ok = true;
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::set_run_DSD(bool p_DSD) {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	if (p_DSD) {
		if (m_can_DSD) {
			if (set_format(kASIODSDFormat)) {
				m_run_DSD = true;
				ok = true;
			}
		}
	}
	else {
		if (set_format(kASIOPCMFormat)) {
			m_run_DSD = false;
			ok = true;
		}
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

ASIOIoFormatType asio_device_t::get_format() {
	ASIOIoFormat opt;
	opt.FormatType = kASIOFormatInvalid;
	EnterCriticalSection(&m_cs);
	if (future(kAsioGetIoFormat, &opt)) {
		switch (opt.FormatType) {
		case kASIOPCMFormat:
			break;
		case kASIODSDFormat:
			break;
		}
	}
	else {
		opt.FormatType = kASIOPCMFormat;
	}
	LeaveCriticalSection(&m_cs);
	return opt.FormatType;
}

bool asio_device_t::get_latencies() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->getLatencies(&m_inputLatency, &m_outputLatency);
	if (rc == ASE_OK) {
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::getLatencies(inputLatency = %d, outputLatency = %d) => %s", m_inputLatency, m_outputLatency, asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::get_sample_position(t_uint64& p_position, t_uint64& p_timestamp) {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOSamples sp;
	ASIOTimeStamp ts;
	ASIOError rc = m_pCoASIO->getSamplePosition(&sp, &ts);
	if (rc == ASE_OK) {
		p_position = ((t_uint64)sp.hi << 32) | (t_uint64)sp.lo;
		p_timestamp = ((t_uint64)ts.hi << 32) | (t_uint64) ts.lo;
		ok = true;
	}
	if (m_trace) {
		pfc::string8 str_position;
		pfc::string8 str_timestamp;
		str_position << p_position;
		str_timestamp << p_timestamp;
		console::printf(TRC_INF "IASIO::getSamplePosition(position = %s, timestamp = %s) => %s", str_position.c_str(), str_timestamp.c_str(), asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

unsigned asio_device_t::get_samplerate() {
	ASIOSampleRate samplerate = 0.0;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->getSampleRate(&samplerate);
	if (rc == ASE_OK) {
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::getSampleRate(sampleRate = %d) => %s", (unsigned)samplerate, asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return (unsigned)samplerate;
}

bool asio_device_t::set_samplerate(unsigned p_samplerate) {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->setSampleRate((ASIOSampleRate)p_samplerate);
	if (rc == ASE_OK) {
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::setSampleRate(sampleRate = %d) => %s", p_samplerate, asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::create_buffers(ASIOCallbacks* p_callbacks) {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	if (m_state == ASIO_STATE_INITIALIZED) {
		//set_format(format);
		if (set_samples()) {
			ASIOError rc = m_pCoASIO->createBuffers(m_bufferInfos.get_ptr(), m_bufferInfos.get_count(), m_samples, p_callbacks);
			if (rc == ASE_OK) {
				m_state = ASIO_STATE_PREPARED;
				ok = true;
			}
			if (m_trace) {
				console::printf(TRC_INF "IASIO::createBuffers(numChannels = %d, bufferSize = %d) => %s", m_bufferInfos.get_count(), m_samples, asio_util_t::getASIOError(rc));
				for (unsigned channel = 0; channel < m_bufferInfos.get_count(); channel++) {
					console::printf(TRC_INF "IASIO::ASIOBufferInfo[isInput = %d, channelNum = %d, buffers[0] = 0x%08x, buffers[1] = 0x%08x",
						m_bufferInfos[channel].isInput,
						m_bufferInfos[channel].channelNum,
						m_bufferInfos[channel].buffers[0],
						m_bufferInfos[channel].buffers[1]
					);
				}
			}
		}
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::dispose_buffers() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->disposeBuffers();
	if (rc == ASE_OK) {
		m_state = ASIO_STATE_INITIALIZED;
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::disposeBuffers() => %s", asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::start() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->start();
	if (rc == ASE_OK) {
		m_state = ASIO_STATE_RUNNING;
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::start() => %s", asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::stop() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->stop();
	if (rc == ASE_OK) {
		m_state = ASIO_STATE_PREPARED;
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::stop() => %s", asio_util_t::getASIOError(rc));
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::output_ready() {
	bool ok = false;
	EnterCriticalSection(&m_cs);
	ASIOError rc = m_pCoASIO->outputReady();
	if (rc == ASE_OK) {
		ok = true;
	}
	LeaveCriticalSection(&m_cs);
	return ok;
}

bool asio_device_t::init(void* sysHandle) {
	bool ok = false;
	ASIOBool rc = m_pCoASIO->init(core_api::get_main_window());
	if (rc == ASIOTrue) {
		m_state = ASIO_STATE_INITIALIZED;
		m_pCoASIO->getDriverName(m_name);
		m_version = m_pCoASIO->getDriverVersion();
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::init(sysHandle = 0x%08x) => %s", sysHandle, (rc == ASIOTrue) ? "ASIOTrue" : "ASIOFalse");
		if (ok) {
			console::printf(TRC_INF "IASIO::getDriverName(name = \"%s\")", m_name);
			console::printf(TRC_INF "IASIO::getDriverVersion() => %d", m_version);
		}
	}
	return ok;
}

bool asio_device_t::set_format(ASIOIoFormatType p_format) {
	bool ok = false;
	ASIOIoFormat opt;
	opt.FormatType = p_format;
	if (future(kAsioSetIoFormat, &opt)) {
		ok = true;
	}
	return ok;
}

bool asio_device_t::set_channels() {
	bool ok = false;
	m_channels = 0;
	ASIOError rc = m_pCoASIO->getChannels(&m_numInputChannels, &m_numOutputChannels);
	if (rc == ASE_OK) {
		m_channels = (t_size)m_numOutputChannels;
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::getChannels(numInputChannels = %d, numOutputChannels = %d) => %s", m_numInputChannels, m_numOutputChannels, asio_util_t::getASIOError(rc));
	}
	return ok;
}

bool asio_device_t::set_channelInfos() {
	bool ok = false;
	if (m_can_DSD) {
		if (set_format(kASIODSDFormat)) {
			m_dsd_channelInfos.set_count(m_channels);
			for (unsigned channel = 0; channel < m_channels; channel++) {
				m_dsd_channelInfos[channel].channel = channel;
				m_dsd_channelInfos[channel].isInput = ASIOFalse;
				m_dsd_channelInfos[channel].isActive = ASIOFalse;
				m_dsd_channelInfos[channel].channelGroup = 0;
				m_dsd_channelInfos[channel].type = 0;
				m_dsd_channelInfos[channel].name[0] = '\0';
				ASIOError rc = m_pCoASIO->getChannelInfo(&m_dsd_channelInfos[channel]);
				if (rc == ASE_OK) {
					ok = true;
				}
				if (m_trace) {
					console::printf(TRC_INF "IASIO::getChannelInfo(channel = %d, isInput = %d, isActive = %d, channelGroup = %d, type = %s, name = \"%s\") => %s",
						m_dsd_channelInfos[channel].channel,
						m_dsd_channelInfos[channel].isInput,
						m_dsd_channelInfos[channel].isActive,
						m_dsd_channelInfos[channel].channelGroup,
						asio_util_t::getSampleTypeName(m_dsd_channelInfos[channel].type),
						m_dsd_channelInfos[channel].name,
						asio_util_t::getASIOError(rc)
						);
				}
			}
			set_format(kASIOPCMFormat);
		}
	}
	m_pcm_channelInfos.set_count(m_channels);
	for (unsigned channel = 0; channel < m_channels; channel++) {
		m_pcm_channelInfos[channel].channel = channel;
		m_pcm_channelInfos[channel].isInput = ASIOFalse;
		m_pcm_channelInfos[channel].isActive = ASIOFalse;
		m_pcm_channelInfos[channel].channelGroup = 0;
		m_pcm_channelInfos[channel].type = 0;
		m_pcm_channelInfos[channel].name[0] = '\0';
		ASIOError rc = m_pCoASIO->getChannelInfo(&m_pcm_channelInfos[channel]);
		if (rc == ASE_OK) {
			ok = true;
		}
		if (m_trace) {
			console::printf(TRC_INF "IASIO::getChannelInfo(channel = %d, isInput = %d, isActive = %d, channelGroup = %d, type = %s, name = \"%s\") => %s",
				m_pcm_channelInfos[channel].channel,
				m_pcm_channelInfos[channel].isInput,
				m_pcm_channelInfos[channel].isActive,
				m_pcm_channelInfos[channel].channelGroup,
				asio_util_t::getSampleTypeName(m_pcm_channelInfos[channel].type),
				m_pcm_channelInfos[channel].name,
				asio_util_t::getASIOError(rc)
			);
		}
	}
	return ok;
}

bool asio_device_t::set_samples() {
	bool ok = false;
	m_minSize = 0;
	m_maxSize = 0;
	m_preferredSize = 0;
	m_granularity = 0;
	ASIOError rc = m_pCoASIO->getBufferSize(&m_minSize, &m_maxSize, &m_preferredSize, &m_granularity);
	if (rc == ASE_OK) {
		m_samples = (t_size)m_preferredSize;
		ok = true;
	}
	if (m_trace) {
		console::printf(TRC_INF "IASIO::getBufferSize(minSize = %d, maxSize = %d, preferredSize = %d, granularity = %d) => %s", m_minSize, m_maxSize, m_preferredSize, m_granularity, asio_util_t::getASIOError(rc));
	}
	return ok;
}

bool asio_device_t::future(long selector, void* opt) {
	bool ok = false;
	ASIOError rc = m_pCoASIO->future(selector, opt);
	if (rc == ASE_SUCCESS) {
		ok = true;
	}
	if (m_trace) {
		switch (selector) {
		case kAsioCanDoIoFormat:
		case kAsioGetIoFormat:
		case kAsioSetIoFormat:
			console::printf(TRC_INF "IASIO::future(selector = %s, FormatType = %s) => %s", asio_util_t::getFutureSelectorName(selector), asio_util_t::getIoFormatName(((ASIOIoFormat*)opt)->FormatType), asio_util_t::getASIOError(rc));
			break;
		default:
			console::printf(TRC_INF "IASIO::future(selector = %s, opt = %d) => %s", asio_util_t::getFutureSelectorName(selector), opt, asio_util_t::getASIOError(rc));
			break;
		}
	}
	return ok;
}
